@@ -178,6 +178,13 @@ def cli():
     show_default=True,
     help="Fully Qualified Domain Name that Orchest listens on.",
 )
+@click.option(
+    "--socket-path",
+    "socket_path",  # name for arg
+    default=None,
+    show_default=True,
+    help="The absolute path of the container runtime socket on the node.",
+)
 @click.option(
     "--userdir-pvc-size",
     "userdir_pvc_size",
@@ -208,6 +215,7 @@ def install(
     dev_mode: bool,
     no_argo: bool,
     fqdn: t.Optional[str],
+    socket_path: t.Optional[str],
     userdir_pvc_size: int,
     builder_pvc_size: int,
     registry_pvc_size: int,
@@ -219,6 +227,7 @@ def install(
         dev_mode,
         no_argo,
         fqdn,
+        socket_path,
         userdir_pvc_size,
         builder_pvc_size,
         registry_pvc_size,
@@ -249,7 +258,7 @@ def uninstall(**common_options) -> None:
     is_flag=True,
     default=True,
     show_default=True,
-    help="Watch cluster status changes.",
+    help="Watch status changes until Orchest has updated.",
 )
 @click.option(
     "--dev/--no-dev",
@@ -312,10 +321,18 @@ def update(
     type=click.Choice(cmds.LogLevel),
     help="Log level to set on Orchest services.",
 )
+@click.option(
+    "--socket-path",
+    "socket_path",  # name for arg
+    default=None,
+    show_default=True,
+    help="The absolute path of the container runtime socket on the node.",
+)
 def patch(
     dev: t.Optional[bool],
     cloud: t.Optional[bool],
     log_level: t.Optional[cmds.LogLevel],
+    socket_path: t.Optional[str],
     **common_options,
 ) -> None:
     """Patch the Orchest Cluster.
@@ -326,7 +343,7 @@ def patch(
         orchest patch --dev
 
     """
-    cmds.patch(dev, cloud, log_level, **common_options)
+    cmds.patch(dev, cloud, log_level, socket_path, **common_options)
 
 
 @cli.command(cls=ClickCommonOptionsCmd)
@@ -366,7 +383,17 @@ def version(json_flag: bool, latest_flag: bool, **common_options) -> None:
     show_default=False,
     help="Get output in json.",
 )
-def status(json_flag: bool, **common_options) -> None:
+@click.option(
+    "--wait",
+    "wait_for_status",  # name for arg
+    default=None,
+    show_default=True,
+    type=click.Choice(cmds.ClusterStatus),
+    help="Wait for cluster status to be e.g. Stopped.",
+)
+def status(
+    json_flag: bool, wait_for_status: t.Optional[cmds.ClusterStatus], **common_options
+) -> None:
     """Get Orchest Cluster status.
 
     If invoked with `--json`, then failure to get Orchest Cluster status
@@ -377,7 +404,7 @@ def status(json_flag: bool, **common_options) -> None:
         kubectl -n <namespace> get orchestclusters <cluster-name> -o jsonpath="{.status.message}"
 
     """
-    cmds.status(json_flag, **common_options)
+    cmds.status(json_flag, wait_for_status, **common_options)
 
 
 @cli.command(cls=ClickCommonOptionsCmd)
@@ -387,7 +414,7 @@ def status(json_flag: bool, **common_options) -> None:
     is_flag=True,
     default=True,
     show_default=True,
-    help="Watch status changes.",
+    help="Watch status changes until Orchest is stopped.",
 )
 def stop(watch: bool, **common_options) -> None:
     """Stop Orchest.
@@ -408,7 +435,7 @@ def stop(watch: bool, **common_options) -> None:
     is_flag=True,
     default=True,
     show_default=True,
-    help="Watch status changes.",
+    help="Watch status changes until Orchest is started.",
 )
 def start(watch: bool, **common_options) -> None:
     """Start Orchest.
@@ -427,7 +454,7 @@ def start(watch: bool, **common_options) -> None:
     is_flag=True,
     default=True,
     show_default=True,
-    help="Watch status changes.",
+    help="Watch status changes until Orchest has restarted.",
 )
 def restart(watch: bool, **common_options) -> None:
     """Restart Orchest.
